/*
 * ===========================
 * VDK Builder
 * Version 0.1
 * Revision 0.0
 * November 1998
 * ===========================
 *
 * Copyright (C) 1998,1999 Mario Motta
 * Developed by Mario Motta <mmotta@guest.net>
 *
 * Based on VDK Library 
 * Copyright (C) 1998, Mario Motta <mmotta@guest.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 */
#ifndef _vdkb_ccproj_h_
#define _vdkb_ccproj_h_

#include <vdkb2/vdkb_tree_coll.h>
#include <vdkb2/vdkb_cregex.h>
#include <fstream>
#include <string>
#include <hash_map.h>
#include <set>
#include <list>
#include <time.h>

const int BUFF_SIZE = 1024;

char *strip_comment( char *in, bool *incoment);

struct hash_str
{
  size_t operator()( const string& str) const
  {
    hash<const char *> hasher;
    return hasher( str.c_str());
  }
};

class file_context
{
  string file_name;
  int linenum;
  char buff[BUFF_SIZE];
 public:
  bool incomment;
  file_context(){ linenum = -1; incomment = false;}
  void set_file( const char *name)
    { file_name = name; linenum = -1; incomment = false;}
  int operator++()
    {
      linenum++;
      return linenum;
    }
  char * fgets( char *s, int size, istream &str)
    {
      str.getline( s, size);
      if (str)
	linenum++;
      return s;
    }
  char *fgets( istream &str)
    {
      str.getline( buff, BUFF_SIZE);
      if (str)
	linenum++;
      return buff;
    }
  int get_linenum(){ return linenum;}
  const char *get_filename() const { return file_name.c_str();}
};

enum fnc_flags
{
  fpublic = 0,
  fprivate = 0x1,
  fprotected = 0x2,
  fstatic = 0x4,
fvirtual = 0x8
};

class VDKBMemberFunction
{
  string declfile;
  int declline;
  string deffile;
  int defline;
  unsigned int flags;
  string name;
 public:
  void setflag( unsigned int flag);
  VDKBMemberFunction()
    {
      declline = -1;
      defline = -1;
      flags = 0;
    }
  VDKBMemberFunction( const VDKBMemberFunction& in)
    {
      *this = in;
    }
  VDKBMemberFunction& operator=( const VDKBMemberFunction& in);
  int getdeclline() const { return declline+1;}
  void setdeclline( int linein){ declline = linein;}
  int getdefline() const { return defline+1;}
  void setdefline( int linein){ defline = linein;}
  const char *getname() const { return name.c_str();}
  void setname( const char *namein){ name = namein;}
  void setdeclfile( const char *namein) { declfile = namein;}
  const char *getdeclfile() const { return declfile.c_str();}
  void setdeffile( const char *namein){ deffile = namein;}
  const char *getdeffile() const  { return deffile.c_str();}
  bool is_defined() const { return defline != -1;}
  unsigned int get_flags() const { return flags;}
};


class VDKBCCProject;

typedef hash_map<string,VDKBMemberFunction,hash_str> funct_map;
typedef funct_map::const_iterator funct_pos;

class VDKBCCClass
{
protected:
    funct_map  functions;
    vector<string> parents;
    static executed_regex forward_dec_re;
    static executed_regex function_dec_re;
    string name;
    string declfile;
    int declline;
    int endline;
    static char scratch[BUFF_SIZE];
    char * get_parents( char *ptr, istream& str, file_context& context);
    char *find_start( char *buff, istream & str, file_context& context);
    void add_function_dec( const char *funname, const char *finame, int lnum,
                          unsigned int flag);
    VDKBCCProject *owner;
public:
    VDKBCCClass(){ owner = NULL;}
    ~VDKBCCClass() {}
    VDKBCCClass( const VDKBCCClass& in)
    {
        *this = in;
    }
    bool is_defined() const { return !declfile.empty();}
    const char *getname() const { return name.c_str();}
    void setname( const char *namein){ name = namein;}
    void setdeclfile( const char *namein){ declfile = namein;}
    const char *getdeclfile() const { return declfile.c_str();}
    const char *getdeffile() const;
    int getdeclline() const { return declline+1;}
    VDKBCCClass& operator=( const VDKBCCClass& in);
    vector<string> get_parents() { return parents;}
    void add_function_def( const char *funname, const char *finame, int lnum);
    void scan_def( char *buff, istream &str, file_context &context,
		   bool is_struct);
    void scan_gui_def( char *buff, istream &str, file_context &context,
		       bool is_struct);
    bool check_for_gui_include(char *buff);
    void setowner( VDKBCCProject *in){ owner = in;}
    funct_pos get_first_function() const { return functions.begin();}
    funct_pos get_last_function() const { return functions.end();}
    const VDKBMemberFunction& get_function( funct_pos in) const
      { return (*in).second;}
    const VDKBMemberFunction& get_function_from_name( const char *nin)
      { return functions[nin];}
    const char* get_file_from_function( const char *func);
    int get_line_from_function( const char *func);
    const char* get_decl_file_from_function( const char *func);
    int get_decl_line_from_function( const char *func);
    int get_new_decl_line( const char *section, bool *section_exists) const;
    int get_new_def_line() const;
};

    

typedef hash_map<string,VDKBCCClass,hash_str> class_map;
typedef class_map::const_iterator class_pos;
typedef list<string>::const_iterator file_pos;
typedef list<tree_coll<string> >::const_iterator dir_pos;
typedef hash_map<string,time_t,hash_str> file_time_map;
typedef const tree_coll<string> & dirref;

struct ltstr
{
  bool operator()(const char* s1, const char* s2) const
  {
    return strcmp(s1, s2) < 0;
  }
};

// forward declaration only
class VDKBProjectManager;

class VDKBCCProject
{
    VDKBProjectManager* owner;
    list<string> files;
    class_map classes;
    string directory;
    static executed_regex memberfunctregex;
    static executed_regex classregex;
    static executed_regex structregex;
    bool scan_file( const char *fname);
    bool check_for_include( char *buff);
    bool check_for_class( char *buff, istream &str, file_context & context);
    bool regex_check_for_class( char *buff, executed_regex & er, istream &str,
				file_context & context, bool is_struct);
    bool check_for_member_function( char *buff, istream &str,
				    file_context & context);
    class_map::iterator get_class( const char *cname);
    void strip_undefined();
public:
    static executed_regex includeregex;
    file_time_map ftimes;
    VDKBCCProject(VDKBProjectManager* own);
    ~VDKBCCProject() {}
    void Clean();
    bool  process_line( char *buff, istream& str, file_context& context);
    bool isupdaterequired();
    void setdirectory( const char *name);
    const char *getdirectory() const { return directory.c_str();}
    bool update();
    class_pos get_first_class() const { return classes.begin();}
    class_pos get_end_class() const { return classes.end();}
    const VDKBCCClass& get_class( class_pos in) const { return (*in).second;}
    VDKBCCClass& get_class_from_name( const char *n);
    char* get_file_from_function( const char *cl, const char *func);
    int get_line_from_function( const char *cl, const char *func);
    char *get_decl_file_from_function( const char *cl, const char *func);
    int get_decl_line_from_function( const char *cl, const char *func);
    char *get_file_from_class( const char *cl);
    int get_line_from_class( const char *cl);
    string getfullname( const char *fname) const;
    
protected:
    bool extract_function( executed_regex& re, file_context &context);
    bool get_file_list( list<string>& flist);
    bool do_update();
    bool check_update_required();
};

#endif // _vdkb_ccproj_h_
