/*
 * ===========================
 * VDK Builder
 * Version 0.1
 * Revision 0.0
 * November 1998
 * ===========================
 *
 * Copyright (C) 1998,1999 Mario Motta
 * Developed by Mario Motta <mmotta@guest.net>
 *
 * Based on VDK Library
 * Copyright (C) 1998, Mario Motta
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#if USE_XDB
#include <vdkxdb2/vdkxdb.h>
#include <vdkxdb2/vdkxtable.h>
#endif
#if !HAVE_GNOME
#if ENABLE_NLS
#include <libintl.h>
//#define _(str) gettext(str)
#define _(str) \
    ( g_utf8_validate(gettext(str),-1,NULL) ? \
    gettext(str) : \
    g_locale_to_utf8(gettext(str),-1,NULL,NULL,NULL) )
#define N_(str) str
#else
#define _(str) str
#define N_(str) str 
#endif
#endif

#define VERBOSE 0
#undef __REGEXP_LIBRARY_H__    // do not include builtin regex.h
#define __REGEXP_LIBRARY_H__ 1 // use it from libc6-dev package
#include <vdkb2/vdkb_ccproj.h>
#include <vdkb2/vdkb_prjman.h>
#include <vdkb2/vdkb_utils.h>
#include <glob.h>
#include <string.h>
#include <algorithm>
#include <unistd.h>
#include <sys/stat.h>
#include <dirent.h>

executed_regex VDKBCCProject::includeregex("^#include +<(.*)>",REG_EXTENDED,1);
executed_regex VDKBCCProject::memberfunctregex( "[ \t]*([^ :]+) *:: *([^ \\(]+) *\\(",
                                            REG_EXTENDED,2);
executed_regex VDKBCCProject::classregex("^[ \t]*class[ \t]+([^ \t:,\n]+)"
                                     ,REG_EXTENDED,1);
executed_regex VDKBCCProject::structregex("^[ \t]*struct[ \t]+([^ \t:,\n]+)",
                                      REG_EXTENDED,1);

char VDKBCCClass::scratch[]="";
executed_regex VDKBCCClass::forward_dec_re(";[ \t]*$",REG_EXTENDED,0);
executed_regex VDKBCCClass::function_dec_re("([^\({:;\\*\\.\\\\/ \t\n&\\!]+) *\\(",
                                        REG_EXTENDED,1);


bool does_file_exist(const char *fname)
{
    ifstream f(fname);
    bool res=true;
    if(!f)
    {
        res=false;
    }
    return res;
}


template <class InputIterator>
int get_bracelevel(InputIterator first,InputIterator last)
{

    int res=0;
    while(first != last)
    {
        if(*first == '{') res ++;
        else if(*first == '}') res --;
        first++;
    }
    return res;
}

char *strip_strings(char *in)
{
    char *ptr;
    char *startstring=strchr(in,'\"');
    while(startstring != NULL)
    {
        ptr=strchr(startstring+1,'\"');
        if(ptr !=NULL)
        {
            memmove(startstring,ptr+1,strlen(ptr+1)+1);
            startstring=strchr(in,'\"');
        }
        else
        {
            *startstring='\0';
            startstring=NULL;
        }
    }
    return in;
}    

char * strip_comment(char *in,bool *incomment)
{
    char *res=in;
    strip_strings(in);
    char *ptr=strstr(in,"//"); // first strip C++ style
    if(ptr != NULL)
    {
        *ptr='\0';
    }
        /* now look at C style */
    if(*incomment)
    {
        res=strstr(in,"*/");
        if(res==NULL)
        {
            res=in + strlen(in);
            return res; // still in commaent
        }
        *incomment=false;
    }
    char *commstart;
    //    char *commend;
    do
    {
        commstart= strstr(res,"/*");
        if(commstart !=NULL)
        {
            ptr=strstr(commstart,"*/");
            if(ptr==NULL)
            {
                *incomment=true;
                *commstart='\0';
                return res;
            }
            memmove(commstart,ptr+2,strlen(ptr+2)+1);
        }
    } while(commstart != NULL);
    return res;
}

void
VDKBMemberFunction::setflag(unsigned int flag)
{
    if(flag == fpublic || flag==fprivate || flag== fprotected)
    {
        flags &= 0x0c;
    }
    flags |= flag;
}

VDKBMemberFunction&
VDKBMemberFunction::operator=(const VDKBMemberFunction& in)
{
    name=in.name;
    declfile=in.declfile;
    declline=in.declline;
    deffile=in.deffile;
    defline=in.defline;
    flags=in.flags;
    return *this;
}

VDKBCCClass&
VDKBCCClass::operator=(const VDKBCCClass& in)
{
    name =in.name;
    declfile=in.declfile;
    declline=in.declline;
    endline=in.endline;
    functions=in.functions;
    owner=in.owner;
    return *this;
}

char*
VDKBCCClass::get_parents(char *ptr,istream &str,file_context& context)
{
    char *scr_ptr;
    scr_ptr=strchr(ptr,':');
    if(scr_ptr!=NULL)
    {
        ptr=scr_ptr;
        ptr++;
        bool done=false;
        bool parent_flag=false;
	//        size_t offset=0;
        while(!done)
        {
            strncpy(scratch,ptr,BUFF_SIZE);
            scr_ptr=strpbrk(scratch,"():{};");
            if(scr_ptr != NULL)
            {
                done=true;
                *scr_ptr='\0';
                ptr += strlen(scratch);
            }
            scr_ptr=strtok(scratch," \t,");
            if(scr_ptr != NULL)
            {
                if(parent_flag || strcmp(scr_ptr,"public")==0 ||
                   strcmp(scr_ptr,"private")==0)
                {
                    parent_flag=true;
                    while(scr_ptr != NULL)
                    {
                        scr_ptr=strtok(NULL," \t,\n");
                        if(scr_ptr !=NULL)
                        {
                            parents.push_back(scr_ptr);
                        }
                    }
                }
            }
            if(!done)
            {
                ptr=context.fgets(str);
                if(!str)return NULL;
                if(owner != NULL && owner->process_line(ptr,str,context)) done=true;
            }
        }
    }
    return ptr;
}

const char*
VDKBCCClass::getdeffile() const
{
    string res;
    funct_pos ptr=get_first_function();
    while(ptr != get_last_function())
    {
        res=(*ptr).second.getdeffile();
        if(res != declfile)
        {
            break;
        }
        ptr++;
    }
    if(res.empty())
    {
        res=declfile;
    }
    return res.c_str();
}

int
VDKBCCClass::get_new_def_line() const
{
    int line=0;
    int fline;
    string deffile=getdeffile();
    funct_pos ptr=get_first_function();
    while(ptr!=get_last_function())
    {
        fline=(*ptr).second.getdefline();
        if(deffile==(*ptr).second.getdeffile())
        {
            line=max(line,fline);
        }       
        ptr++;
    }
    return line-1;
}
        
    
int
VDKBCCClass::get_new_decl_line(const char *section,bool *section_exists)const
{
    char buff[BUFF_SIZE];
    string sect_tag=section;
    sect_tag += ":";
    ifstream f(declfile.c_str());
    int res=-1;
    int i;
    //    char *ptr;
    *section_exists=false;
    if(f)
    {
        for(i=0;i<=declline;i++)
        {
            f.getline(buff,BUFF_SIZE);
        }
        while(i <= endline && f)
        {
            f.getline(buff,BUFF_SIZE);
            if(strstr(buff,sect_tag.c_str())!= NULL)
            {
                res=i+1;
                *section_exists=true;
                break;
            }
            i++;
        }
    }
    if(*section_exists==false)
    {
        res=endline;
    }
    return res;
}        
    
char*
VDKBCCClass::find_start(char *buff,istream & str,file_context& context)
{
    char *res=strchr(buff,'{');
    while(res==NULL && str)
    {
        buff=context.fgets(str);
        if(owner != NULL && owner->process_line(buff,str,context))
        {
            return NULL;
        }
        res=strchr(buff,'{');
    }
    return res;
}

void
VDKBCCClass::scan_def(char *buff,istream &str,file_context& context,bool is_struct)
{
    if(forward_dec_re.regexec(buff)==0)
        return; // forward declaration
    declfile=context.get_filename();
    declline=context.get_linenum();
    char *ptr=strstr(buff,name.c_str());
    ptr += name.size();
    ptr=get_parents(ptr,str,context);
    ptr=find_start(ptr,str,context);
    int brace_depth=0;
    if (ptr != NULL) brace_depth=1;
    ptr++;
    char *name_ptr;
    unsigned int curflag;
    if(is_struct)
    {
        curflag = fpublic;
    }
    else
    {
        curflag = fprivate;
    }
    while(brace_depth > 0 && str)
    {
        if(function_dec_re.regexec(ptr)==0)
        {
            name_ptr=strstr(ptr,(*function_dec_re.begin()).c_str());
            brace_depth += get_bracelevel(ptr,name_ptr);
            if(brace_depth==1)
            {
               add_function_dec( (*function_dec_re.begin()).c_str(),
                                 context.get_filename(),context.get_linenum(),
                                 curflag);
            }
            brace_depth += get_bracelevel(name_ptr,ptr+strlen(ptr));
        }
        else
	  {
	    if (strstr(ptr,"private:")) curflag = fprivate;
	    else if(strstr(ptr,"protected:")) curflag = fprotected;
	    else if(strstr(ptr,"public:")) curflag = fpublic;
	    else if(strstr(ptr,"#include")) check_for_gui_include(ptr);
	    brace_depth += get_bracelevel(ptr,ptr+strlen(ptr));
	  }
        if(brace_depth > 0)
        {
            ptr=context.fgets(str);
//             if(owner != NULL && owner->process_line(ptr,str,context))
// 	      brace_depth=0;
        }
    }
    endline=context.get_linenum();
}

void
VDKBCCClass::scan_gui_def(char *buff,istream &str,file_context& context,bool is_struct)
{
  char *ptr=buff;
  int brace_depth=0;
  if (ptr != NULL) brace_depth=1;

    char *name_ptr;
    unsigned int curflag;
    if(is_struct)
    {
        curflag = fpublic;
    }
    else
    {
        curflag = fprivate;
    }
    while(brace_depth > 0 && str)
    {
        if(function_dec_re.regexec(ptr)==0)
        {
            name_ptr=strstr(ptr,(*function_dec_re.begin()).c_str());
            brace_depth += get_bracelevel(ptr,name_ptr);
            if(brace_depth==1)
            {
               add_function_dec( (*function_dec_re.begin()).c_str(),
                                 context.get_filename(),context.get_linenum(),
                                 curflag);
            }
            brace_depth += get_bracelevel(name_ptr,ptr+strlen(ptr));
        }
        else
	  {
	    if (strstr(ptr,"private:")) curflag = fprivate;
	    else if(strstr(ptr,"protected:")) curflag = fprotected;
	    else if(strstr(ptr,"public:")) curflag = fpublic;
	    else if(strstr(ptr,"#include")) check_for_gui_include(ptr);
	    brace_depth += get_bracelevel(ptr,ptr+strlen(ptr));
	  }
        if(brace_depth > 0)
        {
            ptr=context.fgets(str);
            if(owner != NULL && owner->process_line(ptr,str,context))
	      brace_depth=0;
        }
    }
}

void
VDKBCCClass::add_function_dec(const char *funname,const char *finame,int lnum,
                               unsigned int flag)
{
    funct_map::iterator ptr=functions.find(funname);
    if(ptr==functions.end())
    {
        functions[funname].setname(funname);
        ptr=functions.find(funname);
        if(ptr==functions.end())
            return;
    }
    (*ptr).second.setdeclfile(finame);
    (*ptr).second.setdeclline(lnum);
    if((*ptr).second.is_defined()==false)
    {
        (*ptr).second.setdeffile(finame);
        (*ptr).second.setdefline(lnum);
    }
    (*ptr).second.setflag(flag);
}
        

void
VDKBCCClass::add_function_def(const char *funname,const char *finame,int lnum)
{
    funct_map::iterator ptr=functions.find(funname);
    if(ptr==functions.end())
    {
        functions[funname].setname(funname);
        ptr=functions.find(funname);
        if(ptr==functions.end())
            return;
    }
    (*ptr).second.setdefline(lnum);
    (*ptr).second.setdeffile(finame);
}

const char*
VDKBCCClass::get_file_from_function(const char *func) 
{
    return functions[func].getdeffile();
}

int
VDKBCCClass::get_line_from_function(const char *func) 
{
    return functions[func].getdefline();
}

const char*
VDKBCCClass::get_decl_file_from_function(const char *func) 
{
    return functions[func].getdeclfile();
}

int
VDKBCCClass::get_decl_line_from_function(const char *func) 
{
    return functions[func].getdeclline();
}
    
void
VDKBCCProject::strip_undefined()
{
    class_map::iterator ptr,tmp_ptr;
    ptr=classes.begin();
    while(ptr != classes.end())
    {
        if((*ptr).second.is_defined()==false)
        {
            tmp_ptr=ptr++;
            classes.erase(tmp_ptr);
        }
        else
            ptr++;
    }
}
            
class_map::iterator
VDKBCCProject::get_class(const char *cname)
{
    class_map::iterator res=classes.find(cname);
    if(res==classes.end())
    {
        classes[cname].setname(cname);
        res=classes.find(cname);
        (*res).second.setowner(this);
    }
    return res;
}

bool
VDKBCCProject::regex_check_for_class(char *buff,executed_regex & er,
                                      istream &str,file_context & context,
                                      bool is_struct)
{
    bool res=false;
    if(er.regexec(buff)==0)
    {
        string class_str= *(er.begin());
        if(class_str.length() > 0)
        {
            (*get_class(class_str.c_str())).second.scan_def(buff,str,context,
                                                            is_struct);
            res=true;
        }
    }
    return res;
}

VDKBCCClass&
VDKBCCProject::get_class_from_name(const char *n) 
{
    return classes[n];
}
    
bool
VDKBCCProject::check_for_class(char *buff,istream& str,file_context & context)
{
    bool res=regex_check_for_class(buff,classregex,str,context,false);
    if(!res) res=regex_check_for_class(buff,structregex,str,context,true);
    return res;
}

bool
VDKBCCProject::check_for_member_function(char *buff,istream &str,
                                          file_context & context)
{
    bool res=false;
    //    vector<string>::const_iterator ptr;
    if(memberfunctregex.regexec(buff)==0)
    {
        res=extract_function(memberfunctregex,context);
    }
    return res;
}

bool
VDKBCCProject::extract_function(executed_regex& re,file_context &context)
{
    bool res=false;
    string class_str,funct_str;
    vector<string>::const_iterator ptr=re.begin();
    class_str= *ptr;
    ptr++;
    funct_str=*ptr;
    if(funct_str.length() > 0)
    {
        (*get_class(class_str.c_str())).second.
            add_function_def(funct_str.c_str(),
                             context.get_filename(),context.get_linenum());
        res=true;
    }
	return res;
}
    
        
bool
VDKBCCProject::process_line(char *buff,istream& str,file_context &context)
{
    bool res=check_for_include(buff);
    if(!res)
    {
        strip_comment(buff,&context.incomment);
        res=check_for_class(buff,str,context);
        if(!res)
        {
            res=check_for_member_function(buff,str,context);
        }
    }
    return res;
}

bool
VDKBCCProject::isupdaterequired()
{
    bool res=false;
    if(directory.empty()==false)
    {
        chdir(directory.c_str());
        res = check_update_required();
    }
    return res;
}

bool
VDKBCCProject::check_update_required()
{
    bool res=false;;
    list<string> flist;
    list<string>::iterator fptr;
    struct stat fstat;

    flist.erase(flist.begin(),flist.end());
    if(get_file_list(flist))
    {
	files = flist;
        fptr=flist.begin();
        while(fptr != flist.end() && res==false)
        {
            if(ftimes.count((*fptr).c_str())==0)
            {
                if(does_file_exist((*fptr).c_str()))
                res=true;
            }
            else
            {
                stat((*fptr).c_str(),&fstat);
                if(ftimes[(*fptr).c_str()] < fstat.st_mtime)
                {
                    res=true;
                }
            }
            fptr++;
        }
    }
    return res;
}
                
            
            

void
VDKBCCProject::setdirectory(const char *name)
{
    directory=name;
    if(directory[directory.length()-1] != '/')
        directory += '/';
}

bool
VDKBCCProject::get_file_list(list<string> &flist)
{ // here scans for files in project
#if VERBOSE
  g_print("Getting file list...\n");
#endif
  UnitListIterator li(owner->Project()->Units());

  int type;
  char* fname;
  for (;li;li++)
  {
	type = li.current()->Type();
	//	g_print("Taken type %d\n",type);
	if ( type == source_unit || type == c_source_unit || type == header_unit || type == project_unit)
	{
	  if(does_file_exist(fname=li.current()->Source())) {
		flist.push_back(string(fname));
		// g_print("Added ... %s\n",fname);
	  }
	  if(does_file_exist(fname=li.current()->Header())) {
		flist.push_back(string(fname));
		// g_print("Added ... %s\n",fname);
	  }
	  if(does_file_exist(fname=li.current()->GuiSource())) {
		flist.push_back(string(fname));
		// g_print("Added ... %s\n",fname);
	  }
	  if(does_file_exist(fname=li.current()->DfmHeader())) {
		flist.push_back(string(fname));
		// g_print("Added ... %s\n",fname);
	  }
	}
  }

  return true;
}

bool
VDKBCCProject::do_update()
{
    bool res;
    list<string>::iterator fptr;

    res=get_file_list(files);
    if(res)
    {
      classes.erase(classes.begin(),classes.end());
        fptr=files.begin();
        while(fptr!=files.end())
        {
            scan_file((*fptr).c_str());
            fptr++;
        }
    }
    return res;
}

                

bool
VDKBCCProject::update()
{
    bool res;
    files.erase(files.begin(),files.end());
    if(chdir(directory.c_str()) < 0)
    {
#if VERBOSE
        cerr<<"error changing to directory "<<directory<<endl;
#endif
        return false;
    }
    if(!includeregex.verify_regex(cerr) ||
       !memberfunctregex.verify_regex(cerr)   )
    {
        return false;
    }
    res=do_update();
    strip_undefined();
    return res;
}

bool
VDKBCCProject::check_for_include(char *buff)
{
    bool res=false;
    if(includeregex.regexec(buff)==0)
    {
        string fname= *(includeregex.begin());
        if(fname.length() > 0)
        {
            scan_file(fname.c_str());
            res=true;
        }
    }
    return res;
}

bool
VDKBCCClass::check_for_gui_include(char *buff)
{
    bool res=false;
    if(owner->includeregex.regexec(buff)==0)
    {
        string fname= *(owner->includeregex.begin());
        if(fname.length() > 0)
        {
	  ifstream fptr;
	  string fullname;
	  struct stat fstat;
	  char *ptr;
	  //    bool oldfile=false;
	  file_context context;
	  
	  fullname=owner->getfullname(fname.c_str());
	  stat(fullname.c_str(),&fstat);
	  fptr.open(fullname.c_str());
	  if(!fptr)
	    {
#if VERBOSE
	      cerr<<"check_for_gui_include() - Error opening file "<<fullname<<endl;
#endif
	      return false;
	    }
	  context.set_file(fullname.c_str());
	  while(fptr)
	    {
	      ptr=context.fgets(fptr);
	      scan_gui_def(ptr,fptr,context, false);
	    }
	  owner->ftimes[fullname]=fstat.st_mtime;
	  res=true;
        }
    }
    return res;
}

string
VDKBCCProject::getfullname(const char *fname) const
{
    string res=fname;
    list<string> flist = files;
    list<string>::iterator fptr;

	if ( *fname != '/')
	{
	  fptr=flist.begin();
	  for(;fptr!=flist.end();fptr++)
	    if ( strstr(fname,(*fptr).c_str()) != NULL )
	      {
		res = (*fptr).c_str();
		return res;
	      }
	  res.insert(0,directory);
	}
    return res;
}
    
bool
VDKBCCProject::scan_file(const char *fname)
{
    ifstream fptr;
    string fullname;
    struct stat fstat;
    char *ptr;
    //    bool oldfile=false;
    file_context context;
    fullname=getfullname(fname);
    stat(fullname.c_str(),&fstat);
    fptr.open(fullname.c_str());
    if(!fptr)
      {
#if VERBOSE
	cerr<<"scan_file() - Error opening file "<<fullname<<endl;
#endif
	return false;
      }
#if VERBOSE
	cerr<<"scan_file() - file opened: "<<fullname<<endl;
#endif
    context.set_file(fullname.c_str());
    while(fptr)
      {
	ptr=context.fgets(fptr);
	process_line(ptr,fptr,context);
      }
    ftimes[fullname]=fstat.st_mtime;
    return true;
}

char*
VDKBCCProject::get_file_from_class(const char *cl)
{
  /*
    string name(classes[cl].getdeclfile());
    return name.c_str();
  */
  char* result = NULL;
  const char* name = classes[cl].getdeclfile();
  if(name)
    {
      result = new char[strlen(name)+1];
      strcpy(result,name);
    }
  return result;
}

int
VDKBCCProject::get_line_from_class(const char *cl)
{
    return classes[cl].getdeclline();
}

char*
VDKBCCProject::get_file_from_function(const char *cl,const char *func)
{
  /*
    string name(classes[cl].get_file_from_function(func));
    return name.c_str();
  */
  char* result = NULL;
  const char* name = classes[cl].get_file_from_function(func);
  if(name)
    {
      result = new char[strlen(name)+1];
      strcpy(result,name);
    }
  return result;
}

int
VDKBCCProject::get_line_from_function(const char *cl,const char *func) 
{
    return classes[cl].get_line_from_function(func);
}

char*
VDKBCCProject::get_decl_file_from_function(const char *cl,const char *func)
{
  /*
    string name(classes[cl].get_decl_file_from_function(func));
    return name.c_str();
  */
  char* result = NULL;
  const char* name = classes[cl].get_decl_file_from_function(func);
  if(name)
    {
      result = new char[strlen(name)+1];
      strcpy(result,name);
    }
  return result;
}

int
VDKBCCProject::get_decl_line_from_function(const char *cl,const char *func) 
{
    return classes[cl].get_decl_line_from_function(func);
}

void
VDKBCCProject::Clean()
{
  files.erase(files.begin(),files.end());
  classes.erase(classes.begin(),classes.end());
}

VDKBCCProject::VDKBCCProject(VDKBProjectManager* own): owner(own)
{
}
